@@ -8,11 +8,13 @@ from datetime import timedelta
 from logging.config import dictConfig
 from typing import Any, Dict, List, NamedTuple, Optional
 
+from pyspark import SparkContext
 from pyspark.sql import DataFrame, SparkSession, Window
 from pyspark.sql import functions as func
 from pyspark.sql.functions import (
     broadcast,
     col,
+    expr,
     monotonically_increasing_id,
     row_number,
 )
@@ -320,11 +322,16 @@ class FileDestination(NamedTuple):
 
 def _map_column(df: DataFrame, col_mapping: Dict[str, str]):
     source_to_alias_map = {v: k for k, v in col_mapping.items()}
-    projection = [
-        col(col_name).alias(source_to_alias_map.get(col_name, col_name))
-        for col_name in df.columns
-    ]
-    return df.select(projection)
+    projection = {}
+
+    for col_name in df.columns + list(set(col_mapping) - set(df.columns)):
+        if col_name in source_to_alias_map:
+            # column rename
+            projection[source_to_alias_map.get(col_name)] = col(col_name)
+        else:
+            projection[col_name] = expr(col_mapping.get(col_name, col_name))
+
+    return df.select([c.alias(a) for a, c in projection.items()])
 
 
 def as_of_join(
@@ -538,6 +545,9 @@ def join_entity_to_feature_tables(
         joined_df = as_of_join(
             joined_df, entity_event_timestamp_column, feature_table_df, feature_table,
         )
+        if SparkContext._active_spark_context._jsc.sc().getCheckpointDir().nonEmpty():
+            joined_df = joined_df.checkpoint()
+
     return joined_df
 
 
@@ -575,16 +585,18 @@ def filter_feature_table_by_time_range(
 
     time_range_filtered_df = feature_table_df.filter(feature_table_timestamp_filter)
 
+    entities_projected = (
+        entity_df.withColumnRenamed(
+            entity_event_timestamp_column, ENTITY_EVENT_TIMESTAMP_ALIAS
+        )
+        .select(feature_table.entity_names + [ENTITY_EVENT_TIMESTAMP_ALIAS])
+        .distinct()
+    )
+
     time_range_filtered_df = (
         time_range_filtered_df.repartition(200)
         .join(
-            broadcast(
-                entity_df.withColumnRenamed(
-                    entity_event_timestamp_column, ENTITY_EVENT_TIMESTAMP_ALIAS
-                )
-            ),
-            on=feature_table.entity_names,
-            how="inner",
+            broadcast(entities_projected), on=feature_table.entity_names, how="inner",
         )
         .withColumn(
             "distance",
@@ -601,9 +613,9 @@ def filter_feature_table_by_time_range(
             ),
         )
         .where(col("distance") == col("min_distance"))
-        .select(time_range_filtered_df.columns + [ENTITY_EVENT_TIMESTAMP_ALIAS])
-        .localCheckpoint()
     )
+    if SparkContext._active_spark_context._jsc.sc().getCheckpointDir().nonEmpty():
+        time_range_filtered_df = time_range_filtered_df.checkpoint()
 
     return time_range_filtered_df
 
@@ -848,6 +860,7 @@ def _get_args():
     parser.add_argument(
         "--destination", type=str, help="Retrieval result destination in json string"
     )
+    parser.add_argument("--checkpoint", type=str, help="Spark Checkpoint location")
     return parser.parse_args()
 
 
@@ -876,6 +889,9 @@ if __name__ == "__main__":
     feature_tables_sources_conf = json_b64_decode(args.feature_tables_sources)
     entity_source_conf = json_b64_decode(args.entity_source)
     destination_conf = json_b64_decode(args.destination)
+    if args.checkpoint:
+        spark.sparkContext.setCheckpointDir(args.checkpoint)
+
     try:
         start_job(
             spark,
